# -*- coding: utf-8 -*-


import os
import pandas

import unittest
from mockito import mock, unstub

import bigtempo.utils as utils


def generate_for_references(engine, references, symbol, start, end, test_data_dir, module=None):
    for reference in references:
        generate_for_reference(engine, reference, symbol, start, end, test_data_dir, module)


def generate_for_reference(engine, reference, symbol, start, end, test_data_dir, module=None):
    if not os.path.isfile(_get_datafile_path(reference, symbol, test_data_dir)):
        return

    class CustomDatasourceTestCase(_create_datasource_test_case_for(engine)):

        def test_datasource_using_test_data(self):
            _assert_datasource_correctness_using_datafiles(engine, reference, symbol, start, end, test_data_dir)

    CustomDatasourceTestCase.__name__ = '%s{%s}[%s:%s]' % (reference, symbol, start, end)

    if module:
        setattr(module,
                'assert_datasource_correctness_using_datafiles(%s){%s}[%s:%s]' % (reference, symbol, start, end),
                CustomDatasourceTestCase)


def _create_datasource_test_case_for(engine):
    class DatasourceTestCase(unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            testing_builder_mock = mock()

            def testing_builder(cls):
                mock_result = testing_builder_mock.build(cls)
                if mock_result:
                    return mock_result
                return cls()

            cls.testing_builder_mock = testing_builder_mock
            cls.original_builder = engine._builder
            engine._builder = testing_builder

        @classmethod
        def tearDownClass(cls):
            engine._builder = cls.original_builder
            engine._instances = {}

        def setUp(self):
            engine._instances = {}

        def tearDown(self):
            unstub()

    return DatasourceTestCase


def _get_datafile_path(reference, symbol, test_data_dir):
    filename = '%s{%s}.csv' % (reference, symbol)
    return os.path.join(test_data_dir, filename)


def _assert_datasource_correctness_using_datafiles(engine, reference, symbol, start, end, test_data_dir):
    logger = utils.DatasourceLogger()
    mocked_registrations = {}

    for dependency_reference in engine._registrations[reference]['dependencies']:
        mock_data_file = _get_datafile_path(dependency_reference, symbol, test_data_dir)

        if not os.path.isfile(mock_data_file):
            continue

        datasource_mock_cls = _create_mock_datasource(dependency_reference, mock_data_file, logger)

        if not engine._registrations.get(dependency_reference):
            engine._registrations[dependency_reference] = {}

        mocked_registrations[dependency_reference] = engine._registrations[dependency_reference]
        engine._registrations[dependency_reference] = {
            'class': datasource_mock_cls,
            'lookback': 0,
            'dependencies': set()
        }

    expected_data_file = _get_datafile_path(reference, symbol, test_data_dir)

    actual = engine.get(reference).process(symbol, start, end)
    expected = utils.slice(pandas.DataFrame.from_csv(expected_data_file), start, end)

    logger.log('actual', actual)
    logger.log('expected', expected)

    try:
        _assert_dataframe_almost_equal(expected, actual)
    except Exception, e:
        logger.print_summary()
        raise e
    finally:
        for dependency_reference, original_registration in mocked_registrations.items():
            engine._registrations[dependency_reference] = original_registration


def _create_mock_datasource(mock_reference, mock_data_file, logger):
    class DatasourceMock(object):

        def evaluate(self, context, symbol, start=None, end=None):
            data = pandas.DataFrame.from_csv(mock_data_file)
            logger.log(os.path.basename(mock_data_file), data)
            return data

    return DatasourceMock


def _assert_dataframe_almost_equal(expected, actual, margin=0.0000000001):
    tmp = ((expected.dropna() - actual.dropna()).abs() < margin)
    assert tmp.all().all()



# -*- coding: utf-8 -*-


import bigtempo.utils as utils
import bigtempo.defaults as defaults
import bigtempo.tagselection as tagselection


class DatasourceEngine(object):

    def __init__(self,
                 builder=defaults.builder,
                 processingtask_factory=defaults.processingtask_factory):
        self._registrations = {}
        self._instances = {}
        self._tag_selector = tagselection.TagSelector(self.get)
        self._tag_iteration_manager = tagselection.TagManager(self._registrations)
        self._processingtask_factory = processingtask_factory
        self._builder = builder

    def for_each(self, *selection):
        def wrapper(fn):
            self._tag_iteration_manager.register(fn, *selection)
            return fn
        return wrapper

    def for_synched(self, *selection):
        def wrapper(fn):
            self._tag_iteration_manager.register_synched(fn, selection)
            return fn
        return wrapper

    def datasource(self, reference, dependencies=None, lookback=0, tags=None):
        def wrapper(cls):
            self._register_datasource(reference, cls, dependencies=dependencies, lookback=lookback, declared_tags=tags)
            return cls
        return wrapper

    def _register_datasource(self, reference, cls, dependencies=None, lookback=0, declared_tags=None):
        self._instances[reference] = None
        self._registrations[reference] = {
            'class': cls,
            'lookback': lookback,
            'dependencies': set(dependencies) if dependencies else set()
        }

        infered_tags = self._tag_iteration_manager.infere_tags(reference)
        tags = utils.assure_is_valid_set(declared_tags) | utils.assure_is_valid_set(infered_tags)

        self._registrations[reference]['tags'] = tags
        self._tag_selector.register(reference, tags)
        self._tag_iteration_manager.evaluate_new_candidate(reference)

    def select(self, *selectors):
        return self._tag_selector.get(*selectors)

    def tags(self, *references):
        return self._tag_selector.tags(*references)

    def get(self, reference):
        return self._create_processor(reference)

    def _create_processor(self, reference):
        instance = self._lazyload_datasource(reference)
        dependencies = dict((dependency, self._create_processor(dependency)) for dependency in self._registrations[reference]['dependencies'])
        lookback_period = self._registrations[reference]['lookback']
        return self._processingtask_factory(instance, dependencies, lookback_period)

    def _lazyload_datasource(self, reference):
        if not self._instances.get(reference):
            if not self._registrations.get(reference):
                raise KeyError()
            self._instances[reference] = self._builder(self._registrations[reference]['class'])
        return self._instances[reference]



# -*- coding: utf-8 -*-


import pandas
import datetime

import bigtempo.utils as utils


def builder(cls):
    return cls()


def processingtask_factory(instance, dependencies, *args, **kwargs):
    return DataFrameDatasourceTask(instance, dependencies, *args, **kwargs)


class SimpleDatasourceTask(object):

    def __init__(self, instance, dependencies):
        self._instance = instance
        self._dependencies = dependencies

    def process(self, **kwargs):
        context = self._evaluate_datasource_dependencies(**kwargs)
        return self._instance.evaluate(context, **kwargs)

    def _evaluate_datasource_dependencies(self, **kwargs):
        return dict((reference, dependency.process(**kwargs)) for reference, dependency in self._dependencies.iteritems())


class DataFrameDatasourceTask(object):

    def __init__(self, instance, dependencies, lookback_period=0):
        self._instance = instance
        self._dependencies = dependencies
        # TODO: Remove lookback period from this object, read it from the instance instead
        self._lookback_period = lookback_period

    def process(self, symbol, start=None, end=None):
        context = self._create_context_for(symbol, start, end)
        result = self._instance.evaluate(context, symbol, start, end)
        return utils.slice(result, start, end)

    def _create_context_for(self, symbol, start=None, end=None):
        evaluated_dependencies = self._evaluate_datasource_dependencies(symbol, start, end)
        return DatasourceContext(evaluated_dependencies)

    def _evaluate_datasource_dependencies(self, symbol, start=None, end=None):
        result = {}
        for reference, dependency in self._dependencies.iteritems():
            # TODO: Pass freq on to this object, read it from the instance instead
            new_start = None if not start else evaluate_loopback_period(self._lookback_period, start)
            result[reference] = dependency.process(symbol, new_start, end)
        return result


class DatasourceContext(object):

    def __init__(self, dependencies):
        self._dependencies = dependencies

    def dependencies(self, reference=None):
        return self._dependencies.get(reference) if reference else self._dependencies


def evaluate_loopback_period(lookback_period, date, freq='B'):
    if freq == 'B':
        lookback_period = int(lookback_period * 1.05)
    lookback_period += 1

    return relative_period(-lookback_period, date, freq)


def relative_period(periods, date=None, freq='B'):
    business_day = equivalent_business_day() if not date else equivalent_business_day(date)
    #return business_day + (periods * pandas.tseries.offsets.BDay())
    return (pandas.Period(business_day, freq=freq) + periods).to_timestamp()


def equivalent_business_day(date=None):
    if not date:
        date = datetime.datetime.today().replace(hour=0,
                                                 minute=0,
                                                 second=0,
                                                 microsecond=0)

    isoweekday = date.isoweekday()
    return date if isoweekday <= 5 else date - datetime.timedelta(isoweekday - 5)
